import re
import sys

sys.setrecursionlimit(2**12)

from utilities import *

"""
Grammar Rules (syntax loosely based on boost.spirit)
"""

TR = lambda token:TokenRule(token)
RR = lambda pattern:RegexRule(pattern)

def deepcompare(x,y):
    if isinstance(x,dict) and isinstance(y,dict):
        return x==dict(filter(lambda (k,v):k in x and deepcompare(x[k],v),y.items()))
    else:
        return x==y

class RuleException(Exception): pass

class Rule:
    class Undefined: pass
    position = 0
    def __init__(self):
        raise RuleException('Rule cannot be instantiated.')
    def __getitem__(self,index):
        key = None
        forward = identity()
        backward = identity()
        if isinstance(index,slice):
            key = index.start
            forward = index.stop
            backward = index.step
        else:
            key = index
        return AssignmentRule(self,key,forward,backward)
    def __xor__(self,other):
        return ContextRule(self,other[0])
    def __pos__(self):
        return RepetitionRule(self,None)
    def __neg__(self):
        return OptionRule(self)
    def __add__(self,other):
        return RepetitionRule(self,other[0])
    def __mul__(self,other):
        return OptionRepetitionRule(self,other[0])
    def __or__(self,other):
        return SeparationRule([self,other])
    def __lshift__(self,other):
        return ConcatenationRule([self,other])

class EmptyRule(Rule):
    def __init__(self): pass
    def __str__(self): return ''
    def parse(self,string,context={}): self.position = 0
    def compose(self,context,string=[]): pass

class TokenRule(Rule):
    def __init__(self,token): self.token = token
    def __str__(self): return self.token
    def parse(self,string,context={}):
        position = len(self.token % context)
        if (self.token % context) == string[:position]: self.position = position
        else: raise RuleException('Rule "' + str(self) + '" does not apply to:\n' + string)
    def compose(self,context,string=[]):
        string.append(self.token % context)

class RegexRule(Rule):
    def __init__(self, pattern): self.pattern = pattern
    def __str__(self): return self.pattern
    def parse(self,string,context={}):
        try:
            compiled = re.compile('^' + self.pattern % context)
            match = compiled.match(string)
            self.position = match.end()
            return
        except AttributeError:
            raise RuleException('Rule "' + str(self) + '" does not apply to:\n' + string)
    def compose(self,context,string=[]):
        raise RuleException('Rule "' + str(self) + '" does not apply to:\n' + str(context))

class SeparationRule(Rule):
    def __init__(self, rules): self.rules = rules
    def __str__(self):
        try:
            return '(' + '|'.join(map(str,self.rules)) + ')'
        except RuntimeError:
            return ''
    def parse(self,string,context={}):
        for rule in self.rules:
            try:
                rule.parse(string,context)
                self.position = rule.position
                return
            except RuleException: pass
        raise RuleException('Rule "' + str(self) + '" does not apply to:\n' + string)
    def compose(self,context,string=[]):
        result = []
        for rule in self.rules:
            try:
                substring = []
                subcontext = copy.deepcopy(context)
                rule.compose(subcontext,substring)
                result.append((subcontext,substring))
            except RuleException: pass
        if result:
            result.sort(key=lambda x:len(x[0]))
            string.extend(result[0][1])
            return
        raise RuleException('Rule "' + str(self) + '" does not apply to:\n' + str(context))
"""
        for rule in self.rules:
            try:
                substring = []
                subcontext = {}
                rule.compose(context,substring,value)
                result = ''.join(substring)
                if value==Rule.Undefined:
                    rule.parse(result,subcontext)
                    if deepcompare(subcontext,context):
                        string.extend(substring)
                        return
                elif value==result:
                    string.extend(substring)
                    return
            except RuleException: pass
        raise RuleException('Rule "' + str(self) + '" does not apply to:\n' + str(context))
"""
class ConcatenationRule(Rule):
    def __init__(self,rules): self.rules = rules
    def __str__(self):
        try:
            return '(' + ''.join(map(str, self.rules)) + ')'
        except RuntimeError:
            return ''
    def parse(self,string,context={}):
        position = 0
        for rule in self.rules:
            rule.parse(string[position:],context)
            position += rule.position
        self.position = position
    def compose(self,context,string=[]):
        for rule in self.rules:
            rule.compose(context,string)

class AssignmentRule(Rule):
    def __init__(self,rule,key,forward,backward):
        self.rule = rule
        self.key = key
        self.forward = forward
        self.backward = backward
    def __str__(self):
        return '(?P<' + self.key + '>' + str(self.rule) + ')'
    def parse(self,string,context={}):
        self.rule.parse(string, context)
        self.position = self.rule.position
        if not self.key: self.forward(string[:self.position])
        else: context[self.key] = self.forward(string[:self.position])
    def compose(self,context,string=[]):
        try:
            if not self.key: self.backward(context[self.key])
            else: value = self.backward(context[self.key])
            self.rule.parse(value,context)
            string.append(value)
            del context[self.key]
        except (KeyError,RuleException):
            raise RuleException('Rule "' + str(self) + '" does not apply to:\n' + str(context))
"""
        try:
            if not self.key: self.backward(context[self.key])
            else: value = self.backward(context[self.key])
            self.rule.compose(context,string,value)
        except KeyError:
            raise RuleException('Rule "' + str(self) + '" does not apply to:\n' + str(context))
"""

class ContextRule(Rule):
    def __init__(self,rule,context):
        self.rule = rule
        self.context = context
    def __str__(self):
        return '(?C<' + self.context + '>' + str(self.rule) + ')'
    def parse(self, string, context={}):
        try:
            context[self.context] = context.get(self.context,{})
            self.rule.parse(string, context[self.context])
            self.position = self.rule.position
        except RuleException:
            del context[self.context]
            raise
    def compose(self,context,string=[]):
        try:
            self.rule.compose(context[self.context],string)
            del context[self.context]
        except KeyError:
            raise RuleException('Rule "' + str(self) + '" does not apply to:\n' + str(context))

class OptionRule(SeparationRule):
    def __init__(self,rule):
        SeparationRule.__init__(self,[rule,EmptyRule()])
    def __str__(self):
        return str(self.rules[0]) + '?'

class OptionRepetitionRule(SeparationRule):
    def __init__(self,rule,context):
        repetition = self
        if context: repetition = repetition^[context]
        SeparationRule.__init__(self,[rule<<repetition,EmptyRule()])
    def __str__(self):
        return str(self.rules[1]) + '*'

class RepetitionRule(ConcatenationRule):
    def __init__(self,rule,context):
        repetition = OptionRepetitionRule(rule,context)
        if context: repetition = repetition^[context]
        ConcatenationRule.__init__(self,[rule,repetition])
    def __str__(self):
        return str(self.rules[0]) + '+'