from __future__ import division

import math
import numpy

import pygame
from pygame.locals import *

from scene import *
from shape import *
from interpolator import *

class View:
    def __init__(self,viewport):
        self.viewport = viewport
        self.surface = pygame.Surface(self.viewport.size,pygame.SRCALPHA,32)
    def appear(self): pass
    def disappear(self): pass
    def update(self,delta): pass
    def render(self,screen):
        screen.blit(self.surface,self.viewport.topleft)
    def handle(self,event): pass
    def from_viewport(self,point):
        point[1] = self.viewport.size[1]-point[1]
        point *= 2.0
        point[0] /= self.viewport.size[0]
        point[1] /= self.viewport.size[1]
        point -= 1.0
        return point
    def to_viewport(self,point):
        point[1] = -point[1]
        point += 1.0
        point[0] *= self.viewport.size[0]
        point[1] *= self.viewport.size[1]
        point /= 2.0
        return point

class GroupView(View):
    def __init__(self,viewport):
        View.__init__(self,viewport)
        self.views = []
        self.color = (0,0,0)
    def append(self,view):
        self.views.append(view)
    def remove(self,view):
        self.views.remove(view)
    def appear(self):
        for view in self.views:
            view.appear()
        View.appear(self)
    def disappear(self):
        for view in self.views:
            view.disappear()
        View.disappear(self)
    def update(self,delta):
        for view in self.views:
            view.update(delta)
        View.update(self,delta)
    def render(self,screen):
        screen.fill(self.color)
        for view in self.views:
            view.render(screen)
    def handle(self,event):
        for view in self.views:
            view.handle(event)
        View.handle(self,event)

class TextView(View):
    font = None
    def __init__(self,text,color=(255,255,255)):
        if TextView.font is None:
            pygame.font.init()
            TextView.font = pygame.font.Font(None,24)
        self.set(text,color)
        self.viewport = self.surface.get_rect()
    def set(self,text,color=(255,255,255)):
        self.text = text
        self.color = color
        self.surface = TextView.font.render(self.text,1,self.color)

class SceneView(View):
    def __init__(self,viewport):
        View.__init__(self,viewport)
        self.scene = Scene(numpy.array([0,0,0]))
        self.fill = True
        self.stroke = False
        self.light = True
        self.cull = 0 #0 CCW, 1 CW, 2 No Cull
    def render(self,screen):
        def norm_squared(a): return a[0]*a[0]+a[1]*a[1]+a[2]*a[2]
        def cross_z(a,b): return a[0]*b[1]-a[1]*b[0]
        def dot_w(p,pt):
            p = numpy.dot(p,pt)
            p /= p[3]
            return p
        self.surface.fill(self.scene.color)
        camera_transform = numpy.linalg.inv(self.scene.camera.absolute_transform)
        camera_projection = self.scene.camera.project
        camera_transform_nodes = [(numpy.dot(node.absolute_transform,camera_transform),node) for node in self.scene.shape_nodes()]
        #camera_transform_nodes.sort(key=lambda (transform,_):norm_squared(transform[3]),reverse=True)
        camera_transform_nodes.sort(key=lambda (transform,_):transform[3,2],reverse=False)
        transformed_light = self.scene.light #numpy.dot(self.scene.light,camera_transform)
        brightness = 1.0
        for transform, node in camera_transform_nodes:
            direction = (-1,1,0)[self.cull]
            projection_transform = numpy.dot(transform,camera_projection)
            transformed_vertices = [self.to_viewport(dot_w(vertex,projection_transform)) for vertex in node.shape.vertices]
            for i,n in enumerate(node.shape.triangles()):
                if self.light:
                    transformed_normal = numpy.dot(node.shape.normals[i],transform)
                    brightness = max(0.0,numpy.dot(transformed_light,transformed_normal)*direction)
                transformed_triangle = [transformed_vertices[index] for index in n]
                if(cross_z(transformed_triangle[1]-transformed_triangle[0],transformed_triangle[2]-transformed_triangle[0])*direction>=0):
                    pointlist = [(x,y) for x,y,_,_ in transformed_triangle]
                    if self.fill: pygame.draw.polygon(self.surface,node.color*brightness,pointlist,0)
                    if self.stroke: pygame.draw.polygon(self.surface,node.color,pointlist,1)
        View.render(self,screen)
    def update(self,delta):
        self.scene.update(delta)

class TestView(SceneView):
    def __init__(self,viewport):
        SceneView.__init__(self,viewport)
        self.initialize()
    def initialize(self):
        self.scene.color = numpy.array([31,31,63])
        self.scene.light = numpy.array([1.0,0.0,-1.0,0.0])
        self.scene.light /= numpy.linalg.norm(self.scene.light)
        self.scene.camera = CameraNode()
        #aspect = self.viewport.size[1]/self.viewport.size[0]
        #self.scene.camera.project = matrix.ortho(-10.0,10.0,-10.0*aspect,10.0*aspect,0.0,100.0)
        self.scene.camera.project = matrix.perspective(45.0,self.viewport.size[0]/self.viewport.size[1],1.0,100.0)
        self.scene.camera.translation = numpy.array([0.0,1.0,5.0])
        self.scene.root.append(self.scene.camera)
        '''
        self.group = GroupNode()
        self.group.orientational_velocity = quaternion.rotation(10.0,numpy.array([0.0,1.0,0.0]))
        for i in range(-1,2):
            for j in range(-1,2):
                for k in range(-1,2):
                    node = ShapeNode(CuboidShape(),(numpy.array([i,j,k])+1)*255/2)
                    node.translation = numpy.array([i,j,k])*2
                    node.orientational_velocity = quaternion.rotation(10.0,numpy.array([1.0,0.0,1.0]))
                    self.group.append(node)
        self.scene.root.append(self.group)
        '''
        self.logo = ShapeNode(OBJShape('logo.obj'),numpy.array([255,255,255]))
        self.logo.translation = numpy.array([0.0,1.0,0.0])
        self.interpolator = Interpolator(2000,EASE)
        def adjust(t): self.logo.orientation = quaternion.slerp(quaternion.rotation(180.0,numpy.array([0.0,1.0,0.0])),quaternion.rotation(0.0,numpy.array([0.0,1.0,0.0])),t)
        self.interpolator.action = adjust
        self.interpolator.start()
        self.scene.root.append(self.logo)
    def handle(self,event):
        SceneView.handle(self,event)
        if event.type==MOUSEBUTTONDOWN:
            if event.button==4 or event.button==5:
                direction = numpy.dot(numpy.array([0.0,0.0,2.0*(event.button-4.5),0.0]),matrix.orientation(self.scene.camera.orientation))
                self.scene.camera.translation += direction[:3]
        if event.type==MOUSEMOTION:
            if event.buttons[0]==1:
                self.scene.camera.orientation = quaternion.mul(self.scene.camera.orientation,quaternion.rotation(event.rel[0],numpy.array([0.0,1.0,0.0])))
                self.scene.camera.orientation = quaternion.mul(self.scene.camera.orientation,quaternion.rotation(event.rel[1],numpy.array([1.0,0.0,0.0])))
            if event.buttons[2]==1:
                direction = numpy.dot(numpy.array([-event.rel[0],event.rel[1],0.0,0.0])*0.1,matrix.orientation(self.scene.camera.orientation))
                self.scene.camera.translation += direction[:3]
        if event.type==KEYUP:
            if event.key==K_f: self.fill = not self.fill
            if event.key==K_s: self.stroke = not self.stroke
            if event.key==K_l: self.light = not self.light
    def update(self,delta):
        SceneView.update(self,delta)
        self.interpolator.update(delta)